selfDir_uZPPqC := $(dir $(lastword $(MAKEFILE_LIST)))
root := $(selfDir_uZPPqC)..
include $(root)/lib/common.Makefile


##### Configuration

# Command line
log ?= $(if $(test),,1)
profile ?=
quiet ?=
reverify ?= 1
use_staged ?= $(by_col)

# Makefile
exts ?= csv tsv txt xml
test_n ?= 2

##### Vars/functions

# Paths
datasrc := $(notdir $(realpath .))
bin := $(root)/bin
mappings := $(root)/mappings

# Make
SHELL := /bin/bash
selfMake = $(MAKE) --makefile=../input.Makefile
subMake = $(MAKE) $(@:$(root)/%=%) --directory=$(root)
+_ = $(+:_%=)
addBeforeExt = $(basename $(2))$(1)$(suffix $(2))

# OS
os := $(shell uname)
forOs = $(patsubst %,%-$(filter Linux Darwin,$(os)),$(1))

# Formatting
SED = sed -$(if $(filter Darwin,$(os)),E,r)

# System
date = $(shell date +"%Y-%m-%d-%H-%M-%S")

# Terminal
termCols := $(shell tput cols)
esc := '['
reset := $(esc)'0m'
emph := $(esc)'7m '
endEmph := ' '$(reset)

# Commands
MKDIR = mkdir -p
mkdir = $(MKDIR) $(@D)
CP = cp -p
diff = diff --unified=2
diffIgnoreSpace = $(diff) --ignore-space-change
diffVerbose = $(if $(verbose),diff --side-by-side --left-column\
--width=$(termCols),$(diff))

# BIEN commands
sortFilenames = $(shell $(bin)/sort_filenames $(1))
selfMap = $(bin)/cols 0 0
psqlOpts := --set ON_ERROR_STOP=1 --quiet
psqlAsBien := $(bin)/psql_vegbien $(psqlOpts)
# Usage: ($(inDatasrc); cat $(file))|$(psqlCmd)
inDatasrc = echo 'SET search_path TO "$(datasrc)";'

# SVN
addDir = $(if $(wildcard $(1)/),svn add --depth=empty $(1),svn mkdir $(1))
setSvnIgnore = svn propset svn:ignore $(2) $(1)
define addDirWithIgnore
$(addDir)
$(setSvnIgnore)
endef

##### Environment

export PATH := $(bin):$(PATH)

##### General targets

all: _always maps ;

clean: _always
	$(RM) $(all)

remake: _always clean
	+$(selfMake)
# re-run make so that cache of existing files is reset

make_script = ./$< >$@

%/: % _always ;

%: %.make
	$(if $(wildcard $@),,"time" $(make_script))
# Only remake if doesn't exist. This prevents unintentional remaking when the
# make script is newly checked out from svn (which sets the mod time to now) but
# the output is synced externally.
# Can't remove prereq to do this, because it determines when the rule applies.
.PRECIOUS: % # save partial outputs of aborted src make scripts

##### SVN

add: _always
	$(call setSvnIgnore,.,'*')
	$(call addDirWithIgnore,logs,$$'*.log.sql\n*.trace')
	$(call addDirWithIgnore,verify,'*.out')
	$(call addFile,import_order.txt)

# Adds a new table subdir
%/add: _always
	$(call addDirWithIgnore,$*,'*')
	$(call addDirWithIgnore,$*/logs,$$'*.log.sql\n*.trace')

##### Existing maps discovery

sortFile := import_order.txt

tables := $(if $(wildcard $(sortFile)),$(shell cat $(sortFile)))
    # $(shell) replaces "\n" with " "
allSubdirs := $(call wildcard/,*/)
allTables := $(call sortFilenames,$(filter-out _% verify logs,$(allSubdirs:%/=%)))
joinedTables := $(filter-out $(tables),$(allTables))
allTables := $(joinedTables) $(tables)# move joined tables to beginning
ifeq ($(tables),)# none specified in sort file
tables := $(allTables)
endif

viaMap := %/map.csv
directMap := %/VegBIEN.csv
anyMap := $(viaMap) $(directMap) %/unmapped_terms.csv\
%/new_terms.csv

extsFilter := $(addprefix %.,$(exts))
dataOnly = $(filter $(extsFilter),$(1))

anyTest = $*/test.%
srcsOnly = $(filter-out $(anyMap) $(anyTest) %/logs,$(call dataOnly,$(1)))

via := Veg+

coreMap := $(mappings)/VegCore-VegBIEN.csv
dict := $(mappings)/$(via)-VegCore.csv

viaMaps := $(tables:%=%/map.csv)

autogenMaps := $(subst map.,VegBIEN.,$(viaMaps))
directMaps := $(autogenMaps) $(filter-out $(autogenMaps),\
$(wildcard */VegBIEN.csv))

##### Sources

srcs = $(call sortFilenames,$(call srcsOnly,$(wildcard $*/*)))
nonHeaderSrcs = $(filter-out %/header.csv,$(srcs))
isRef = $(if $(nonHeaderSrcs),,1)
    # empty subdir, so references an already-installed staging table
isXml = $(filter %.xml,$(nonHeaderSrcs))
nonXml = $(if $(isXml),,1)
isCsv = $(if $(nonHeaderSrcs),$(if $(isXml),,1))
    # true if $(srcs) non-empty and contains no *.xml
catSrcs = $(bin)/cat$(if $(nonXml),_csv) $(srcs)
withCatSrcs = $(catSrcs:$(bin)/%=$(bin)/with_%) --

# Usage: `make {--silent|-s} inputs/<datasrc>/cat` (don't echo make commands)
cat: $(tables:%=%/cat) _always ;

%/cat: _always
	$(catSrcs)

##### Staging tables installation

dbExports := $(sort $(wildcard *.schema.sql))# schemas first
dbExports += $(sort $(filter-out $(dbExports),$(wildcard *.sql)))# all others
dbExports := $(strip $(dbExports))# += adds extra whitespace
allInstalls := $(if $(dbExports),sql) $(allTables)

install: _always schema $(allInstalls:%=%/install) ;

uninstall: _always rm_schema ;
# rm_schema will also drop all staging tables

reinstall: _always uninstall install ;

schema: _always
	-echo 'CREATE SCHEMA "$(datasrc)";'|$(psqlAsBien)
# ignore errors if schema exists

rm_schema: _always
	echo 'DROP SCHEMA IF EXISTS "$(datasrc)" CASCADE;'|$(psqlAsBien)

logInstall = $(if $(log),$(if $(quiet),&>,2>&1|tee $(2)\
)$(1)logs/install.log.sql)
logInstall* = $(call logInstall,$*/)
logInstall*Add = $(call logInstall,$*/,-a)# append to log

# Must come before %/install to override it
sql/install: $(dbExports)
	($(inDatasrc); cat $+|grep -vF 'SET search_path')|"time" $(psqlAsBien) \
$(logInstall)

# The export must be created with:
# `--compatible=postgresql --add-locks=false --set-charset --no-create-info`
# Must come before `%.sql: _MySQL/%.sql` to override it
%.data.sql: _MySQL/%.data.sql
	$(bin)/my2pg.data <$< >$@

# The export must be created with:
# `--compatible=postgresql --add-locks=false --set-charset`
# Add `--no-data` to create a schema-only export.
%.sql: _MySQL/%.sql
	$(bin)/my2pg <$< >$@

cleanup = (prefix=; . $(bin)/vegbien_dest; env schema=$(datasrc) table=$*\
$(bin)/csv2db) $(logInstall*Add)

define exportHeader
$(cleanup)
echo 'SELECT * FROM "$(datasrc)"."$*" LIMIT 0;'|$(psqlAsBien) \
--no-align --field-separator=, --pset=footer=off >$*/header.csv
endef

# For staging tables which are derived by joining together other staging tables.
%/install %/header.csv: %/create.sql _always
	($(inDatasrc); echo 'CREATE TABLE "$*" AS'; cat $<; echo ';')|"time" \
$(psqlAsBien) --echo-all --set=table='"$*"' $(logInstall*)
	-$(addRowNum)
	$(exportHeader)
# ignore errors if create.sql already added a primary key
addRowNum = echo 'ALTER TABLE "$(datasrc)"."$*"\
ADD COLUMN row_num serial NOT NULL PRIMARY KEY;'|$(psqlAsBien) $(logInstall*Add)

# The joined tables must be suffixed with ".src" to prevent the creation of a
# row_num column, which collides during joins.
isJoinedTable = $(filter %.src,$*)
hasRowNum = $(if $(isJoinedTable),,1)

%/install: _always
	$(if $(isRef),$(exportHeader),$(if $(nonXml),$(import_install_)))
import_install_ = (prefix=; . $(bin)/vegbien_dest; "time" nice -n +5\
env schema=$(datasrc) table=$* has_row_num=$(hasRowNum) $(bin)/csv2db\
$(catSrcs) $(logInstall*))

##### Maps building

# WARNING: You CANNOT make a subdir using `make inputs/<datasrc>/<subdir>/`.
# You must instead make the entire datasource dir: `make inputs/<datasrc>/`

# Maps to (try to) build are added to this
maps :=

srcRoot = $(mappings)/root.sh
mkSrcMap = $(catSrcs)|(. $(srcRoot); env datasrc=$(datasrc) $(bin)/src_map >$@)

# Via maps cleanup
ifneq ($(filter %/.map.csv.last_cleanup,$(MAKECMDGOALS)),)
%/.map.csv.last_cleanup: %/map.csv $(coreMap) $(dict)
	$(bin)/in_place $< $(bin)/canon 1 $(coreMap)
	$(bin)/in_place $< $(bin)/canon 1 $(dict)
	$(bin)/in_place $< $(bin)/translate 1 $(dict)
	touch $@
	+$(selfMake) $(<:%/map.csv=%/unmapped_terms.csv)
	+$(selfMake) $(<:%/map.csv=%/new_terms.csv)
# Include comment column so commented mappings are never removed
else
%/map.csv: _always
	$(if $(wildcard $@),,$(if $(nonXml),$(mkSrcMap)))
	+$(selfMake) $(@:%/map.csv=%/.map.csv.last_cleanup)
endif

%/VegBIEN.csv: %/map.csv $(coreMap)
	<$< $(bin)/cat_cols 1 2|$(bin)/join $(coreMap)|$(bin)/sort_map >$@
maps += $(autogenMaps)

maps: $(maps) _always ;

all += $(maps)

##### Maps validation

# `tail -n +2`: Remove header before running filter_out_ci because filter_out_ci
# only removes the header if it matches the vocabulary's header.
define newTerms
tail -n +2 $<|$(bin)/cols $(1)$(2) >$@
$(bin)/autoremove $@
endef

%/unmapped_terms.csv: %/map.csv $(coreMap)
	$(call newTerms,1,|$(bin)/filter_out_ci 0 $(coreMap))

%/new_terms.csv: %/map.csv $(coreMap) $(dict) %/unmapped_terms.csv
	$(call newTerms,0,$(rmNewTerms))
rmNewTerms = |$(bin)/filter_out_ci 0 $(coreMap)|$(bin)/filter_out_ci 0 $(dict)\
$(if $(wildcard $(word 4,$+)),|$(bin)/filter_out_ci 0 $(word 4,$+))

termsSubdirs := $(tables)

include $(root)/lib/mappings.Makefile

##### External dependencies

$(root)/%: _always
	+$(subMake)
.PRECIOUS: $(root)/% # let ext. dir's Makefile decide whether to delete on error

##### Mapping

+maps = $(filter %/map.csv %/VegBIEN.csv $(mappings)/%,$(+_))
map = $(if $(srcs),$(withCatSrcs) $(root)/map $(+maps),\
$(shell echo Warning: No input file $*/*.{$(exts)} >&2)false)
# need false to run some command, when prefixed by `env ...` below
map2db = $(if $(use_staged),env in_database=vegbien in_schema=$(datasrc)\
in_table=$*) env out_database=vegbien $(map)

##### Import to VegBIEN

profileTest = $(if $(profile),$(if $(test),1))
profileOnly = -env profile_to=/dev/fd/3 $(map2db) 3>&1 1>&2|\
$(bin)/profile_stats /dev/fd/0

log_ = $*/logs/$(if $(n),n=$(n).,)$(date).log.sql
trace = $(log_:.log.sql=.trace)
import = -$(if $(profileTest),$(profileOnly),(set -x; "time" env commit=1\
$(if $(profile),profile_to=$(trace)) $(map2db)) $(if $(log),\
$(if $(n),,&>$(log_)))$(if $(log),$(if $(n), 2>&1|tee -a $(log_))))
# don't abort on import errors, which often relate to invalid input data

import: $(tables:%=%/import) _always ;

%/import: %/VegBIEN.csv _always
	$(import)
# default:
%/import: _always ;

##### Log files from import

logs := $(wildcard */logs/*.log.sql */logs/*.trace)

rm_logs: _always
	$(RM) $(logs)

##### Verification of import

verify: $(tables:%=%/verify) _always ;

%/verify: verify/%.ref verify/%.out _always
	-$(diffVerbose) $(+_)
# don't abort on verification errors, which are expected during development
# default:
%/verify: verify/%.out _always
	$(if $(shell test -e $< && echo t),cat $<)
# don't run if verify/%.out's default do-nothing action was used
# can't use $(wildcard) because it won't recheck file after verify/%.out is run

psqlExport := "time" $(psqlAsBien) --no-align --field-separator=$$'\t'\
--pset=footer=off --pset=null=NULL
verify = $(if $(reverify),$(psqlExport) --set=datasource="'$(datasrc)'" <$< >$@)

verify/%.out: $(mappings)/verify.%.sql _always
	$(verify)
# default:
verify/%.out: _always ;

all += $(wildcard verify/*.out)

%.ref: %.ref.sql
	($(inDatasrc); cat $<)|$(psqlExport) >$@

##### Editing import

rotate: _always
	echo "UPDATE party SET organizationname = organizationname||'.$(date)'\
WHERE organizationname = '$(datasrc)';"|$(psqlAsBien)

rm: _always
	echo "DELETE FROM party WHERE organizationname = '$(datasrc)';"|\
$(psqlAsBien)

##### Testing

testRefOutput = $(subst .by_col,,$(1))
testRef = $(testRefOutput).ref
hasOwnRef = $(filter $@,$(call testRefOutput,$@))
# filter returns non-empty if they are equal

# `rm $@`: Remove outputs of successful tests to reduce clutter
# `$(foreach use_staged...)`: Run with use_staged=1
define runTest
@echo "Testing $(abspath $@)..."
>$@ env test=1 n=$(test_n) $(1) $(foreach use_staged,1,$(map2db))
@(set -x; $(diffIgnoreSpace) $(call testRef,$@) $@) 2>&1 && rm $@ || { e=$$?;\
$(if $(wildcard $(call testRef,$@)),,cat $@;)\
$(if $(hasOwnRef),\
{\
read -p $(emph)'Accept new test output? (y/n)'$(endEmph) REPLY;\
if test "$$REPLY" = y; then\
(set -x; $(MAKE) $@-ok --directory=$(realpath .) --makefile=../input.Makefile);\
exit 0;\
fi;\
};,\
echo $(emph)"Note: The preceding failed test is compared to another test's\
output"$(endEmph);\
echo $(emph)"When it fails, this always indicates a bug"$(endEmph);\
)\
exit $$e;}
endef

tests :=

# Requires staging tables. To create them, run `make inputs/<datasrc>/install`.
# Non-flat-file inputs fall back to mimicking a successful test
%/test.xml: %/VegBIEN.csv _always
	$(if $(nonXml),$(call runTest,by_col=))
tests += %/test.xml

%/test.by_col.xml: %/VegBIEN.csv _always
	-$(if $(nonXml),$(call runTest,by_col=1))
# Don't abort tester if only by-column test fails

# Only run column-based tests if column-based mode enabled, because these tests
# are much slower than the row-based tests for small numbers of rows
ifneq ($(by_col),)
tests += %/test.by_col.xml
endif

testOutputs := $(foreach test,$(tests),$(tables:%=$(test)))

.PRECIOUS: $(testOutputs) # save outputs of failed tests so they can be accepted

test: _always $(testOutputs) ;

all += $(wildcard %/test*.xml)

# Accepts a test output: make <test_output_path>-ok
%-ok: _always
	mv $* $(call testRef,$*)

accept-all: _always
	+yes|$(selfMake) test

##### Documentation

steps = $(selfMake) -s $*/import test=1 by_col=1 verbosity=2 n=100\
2>&1|$(bin)/debug2redmine >$@

%/logs/steps.by_col.log.sql: _always
	+$(steps)
